# -*- coding: utf-8 -*-
from __future__ import unicode_literals

# Django
from django.conf import settings
from django.core.urlresolvers import reverse
from django.test import TestCase

# Local Apps
import factories
from grapevine.settings import grapevine_settings
from grapevine.hipchat.models import HipChat

# SocialProof (will have to change to demo app when this is
#    pulled out of our codebase and put into PIP)
from stories import models as story_models


class HipChatTester(TestCase):
    def setUp(self):
        self.chat_message = HipChat.objects.create()

    def test_message_formats(self):
        hipchat = factories.HipChatFactory()

        message = "<p>Aenean eu leo quam. Pellentesque ornare sem lacinia <i>quam</i> venenatis vestibulum. <strong><i>Integer</i></strong> posuere erat a ante venenatis dapibus posuere velit aliquet.</p>"
        hipchat.message = message

        # default is HTML
        self.assertEquals(hipchat.message_format, HipChat.FORMAT_HTML)
        self.assertEquals(hipchat.message, message)

        # switch it over to text format
        hipchat.message_format = HipChat.FORMAT_TEXT
        self.assertEquals(hipchat.message_format, HipChat.FORMAT_TEXT)
        self.assertEquals(hipchat.message, "Aenean eu leo quam. Pellentesque ornare sem lacinia quam venenatis vestibulum. Integer posuere erat a ante venenatis dapibus posuere velit aliquet.")



from __future__ import unicode_literals

# Django
from django.conf import settings
from django.contrib import admin
from django.core.urlresolvers import reverse

# Local Apps
from grapevine.admin_base import BaseModelAdmin, SendableAdminMixin
from .models import Email, EmailRecipient, EmailBackend, EmailVariable, \
    RawEvent, Event, EmailEvent, UnsubscribedAddress

IS_SUIT_AVAILBLE = "suit" in settings.INSTALLED_APPS


class EmailableAdminMixin(SendableAdminMixin):
    """
    Used for Sendables specifically of the Emailable variety.
    """
    # Used for admin display purposes
    message_type_verbose = "Email"
    if IS_SUIT_AVAILBLE:
        change_form_template = 'admin/suit_change_emailable_form.html'
    else:
        change_form_template = 'admin/change_emailable_form.html'

    def get_test_recipient(self, request, obj_id):
        return request.user.email

    def render_change_form(self, request, context, add=False, change=False, form_url='', obj=None):
        if obj:
            try:
                context['reply_to'] = obj.get_reply_to()
            except ValueError as e:
                context['error_reply_to'] = "ERROR: %s" % (e.args[0],)
            except NotImplementedError:
                context['error_reply_to'] = "ERROR: Could not generate a `reply_to`. Does this template lack a value?"

            try:
                context['from_email'] = obj.get_from_email()
            except ValueError as e:
                context['error_from_email'] = "ERROR: %s" % (e.args[0],)
            except NotImplementedError:
                context['error_from_email'] = "ERROR: Could not generate a `from_email`. Does this template lack a value?"

            try:
                context['subject'] = obj.get_subject()
            except ValueError:
                context['error_subject'] = "ERROR: Could not populate everything in \"%s\"" % (obj.get_raw_subject(),)
            except NotImplementedError:
                context['error_subject'] = "ERROR: Could not generate a subject. Does the template lack a subject?"

        return super(EmailableAdminMixin, self).render_change_form(request, context, add, change, form_url, obj)


class EmailInlineMixin(object):
    extra = 0

    def has_add_permission(self, obj):
        return False

    def has_delete_permission(self, request, obj):
        return False


class EmailVariableInline(EmailInlineMixin, admin.TabularInline):
    model = EmailVariable
    readonly_fields = ['key', 'value']
    verbose_name_plural = 'Variables'


class EmailRecipientInline(EmailInlineMixin, admin.TabularInline):
    model = EmailRecipient
    readonly_fields = ['address', 'domain', 'name', 'type']
    verbose_name_plural = 'Recipients'


class EmailEventInline(EmailInlineMixin, admin.TabularInline):
    model = EmailEvent

    def admin_raw_event(self, obj):
        url = reverse('admin:emails_rawevent_change', args=(obj.raw_event.pk,))
        return '<a href="%s">%s</a>' % (url, obj.raw_event.pk,)
    admin_raw_event.short_description = "Raw Event"
    admin_raw_event.allow_tags = True

    readonly_fields = ['event', 'admin_raw_event', 'happened_at']
    fields = ['admin_raw_event', 'event', 'happened_at']
    verbose_name_plural = 'Events'


class EmailAdmin(BaseModelAdmin):
    inlines = [EmailRecipientInline, EmailVariableInline, EmailEventInline]
    list_display = ['id', 'subject', 'sent_at', 'status', 'is_test']
    list_filter = ['status']
    search_fields = ('=from_email',)

    # Everything is readonly because this table is inherently immutable.
    # It makes no sense to edit the records of that which has already happened.
    readonly_fields = ['subject', 'html_body', 'text_body', 'from_email', 'reply_to', 'type', 'admin_text_full',
        'status', 'sent_at', 'is_test', 'communication_time', 'guid', 'admin_log', 'backend', 'text_excerpt', 'admin_html']

    def admin_html(self, obj):
        return obj.html_body
    admin_html.short_description = 'HTML'
    admin_html.allow_tags = True

    def text_excerpt(self, obj):
        return obj.text_body[:100]
    text_excerpt.short_description = 'Text Excerpt'

    def admin_text_full(self, obj):
        return obj.text_body.replace('\n', '<br>')
    admin_text_full.short_description = 'Text Body'
    admin_text_full.allow_tags = True

    def admin_log(self, obj):
        return '<pre>%s</pre>' % (obj.log,)
    admin_log.short_description = 'Log'
    admin_log.allow_tags = True

    fieldsets = (
        ('Message Quick View', {'fields':
            ('subject', 'admin_html', 'text_excerpt', 'from_email', 'reply_to',)
        },),
        ('Full Message', {
            'classes': ('collapse',),
            'fields': ('html_body', 'admin_text_full', 'admin_log',)
        },),
        ('Status', {'fields':
            ('type', 'status', 'sent_at', 'is_test',)
        },),
        ('Stats', {'fields':
            ('communication_time', 'guid', 'backend',)
        },),
    )


class EmailRecipientAdmin(BaseModelAdmin):
    raw_id_fields = ['email']
    list_display = ['email', 'address', 'type']


class EmailBackendAdmin(BaseModelAdmin):
    list_display = ['id', 'name', 'path', 'username', 'password']
    actions = None

    def has_delete_permission(self, request, obj=None):
        return False


class RawEventAdmin(BaseModelAdmin):

    readonly_fields = ['backend', 'admin_detail_payload', 'processed_on', 'processed_in',
        'is_queued', 'is_broken', 'remote_ip', 'created_at']

    list_display = ['id', 'backend', 'admin_list_payload', 'processed_on', 'processed_in',
        'remote_ip', 'created_at']

    fieldsets = (
        ('Event',
            {'fields': ('backend', 'admin_detail_payload', 'remote_ip',)},
         ),
        ('Status',
            {'fields': ('processed_on', 'processed_in', 'is_queued', 'is_broken', 'created_at',)},
         )
    )

    def admin_list_payload(self, obj):
        payload = obj.payload.replace('\n', '')[:20]
        return payload

    def admin_detail_payload(self, obj):
        return "<pre>%s</pre>" % (obj.payload,)
    admin_detail_payload.short_description = "Payload"
    admin_detail_payload.allow_tags = True


class UnsubscribedAddressAdmin(BaseModelAdmin):
    raw_id_fields = ['email']
    list_display = ['address', 'created_at']


class EventAdmin(BaseModelAdmin):
    list_display = ['name', 'should_stop_sending']


admin.site.register(Email, EmailAdmin)
admin.site.register(EmailRecipient, EmailRecipientAdmin)
admin.site.register(EmailBackend, EmailBackendAdmin)
admin.site.register(RawEvent, RawEventAdmin)
admin.site.register(Event, EventAdmin)
admin.site.register(UnsubscribedAddress, UnsubscribedAddressAdmin)



from __future__ import unicode_literals

# Django
from django.contrib import admin
from django.conf.urls import patterns, url
from django.core.urlresolvers import reverse

# Local Apps
from grapevine.admin import BaseModelAdmin
from .models import HipChat


class HipChatAdmin(BaseModelAdmin):
    pass

admin.site.register(HipChat, HipChatAdmin)



from __future__ import unicode_literals

# 3rd Party
import factory
from factory import fuzzy

# Local Apps
from models import HipChat

class HipChatFactory(factory.django.DjangoModelFactory):
    FACTORY_FOR = HipChat

    from_name = factory.fuzzy.FuzzyText(length=15)
    message = factory.fuzzy.FuzzyText(length=255)

class NotifyingHipChatFactory(HipChatFactory):
    should_notify = HipChat.YES



from __future__ import unicode_literals

# Local Apps
from grapevine import mixins


class EmailSendable(mixins.Emailable, mixins.SendableMixin):
    class Meta(mixins.SendableMixin.Meta):
        abstract = True


class EmailTemplateSendable(mixins.Emailable,
                            mixins.TemplateSendableMixin):
    class Meta(mixins.TemplateSendableMixin.Meta):
        abstract = True



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'HipChat'
        db.create_table(u'hipchat_hipchat', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('created_at', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('updated_at', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
            ('type', self.gf('django.db.models.fields.related.ForeignKey')(default=None, to=orm['contenttypes.ContentType'], null=True, blank=True)),
            ('html_body', self.gf('django.db.models.fields.TextField')(blank=True)),
            ('text_body', self.gf('django.db.models.fields.TextField')(blank=True)),
            ('status', self.gf('django.db.models.fields.IntegerField')(default=2, db_index=True)),
            ('sent_at', self.gf('django.db.models.fields.DateTimeField')(default=None, null=True, db_index=True, blank=True)),
            ('communication_time', self.gf('django.db.models.fields.DecimalField')(default=None, null=True, max_digits=8, decimal_places=5, blank=True)),
            ('is_test', self.gf('django.db.models.fields.BooleanField')(default=False, db_index=True)),
            ('guid', self.gf('django.db.models.fields.CharField')(db_index=True, max_length=36, null=True, blank=True)),
            ('log', self.gf('django.db.models.fields.TextField')(default=None, null=True, blank=True)),
            ('room_id', self.gf('django.db.models.fields.IntegerField')(default=497734)),
            ('color', self.gf('django.db.models.fields.CharField')(default=u'yellow', max_length=50)),
            ('message_format', self.gf('django.db.models.fields.CharField')(default=u'html', max_length=50)),
            ('from_name', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('should_notify', self.gf('django.db.models.fields.IntegerField')(default=1)),
        ))
        db.send_create_signal(u'hipchat', ['HipChat'])


    def backwards(self, orm):
        # Deleting model 'HipChat'
        db.delete_table(u'hipchat_hipchat')


    models = {
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'hipchat.hipchat': {
            'Meta': {'object_name': 'HipChat'},
            'color': ('django.db.models.fields.CharField', [], {'default': "u'yellow'", 'max_length': '50'}),
            'communication_time': ('django.db.models.fields.DecimalField', [], {'default': 'None', 'null': 'True', 'max_digits': '8', 'decimal_places': '5', 'blank': 'True'}),
            'created_at': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'from_name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'guid': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '36', 'null': 'True', 'blank': 'True'}),
            'html_body': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_test': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'db_index': 'True'}),
            'log': ('django.db.models.fields.TextField', [], {'default': 'None', 'null': 'True', 'blank': 'True'}),
            'message_format': ('django.db.models.fields.CharField', [], {'default': "u'html'", 'max_length': '50'}),
            'room_id': ('django.db.models.fields.IntegerField', [], {'default': '497734'}),
            'sent_at': ('django.db.models.fields.DateTimeField', [], {'default': 'None', 'null': 'True', 'db_index': 'True', 'blank': 'True'}),
            'should_notify': ('django.db.models.fields.IntegerField', [], {'default': '1'}),
            'status': ('django.db.models.fields.IntegerField', [], {'default': '2', 'db_index': 'True'}),
            'text_body': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'default': 'None', 'to': u"orm['contenttypes.ContentType']", 'null': 'True', 'blank': 'True'}),
            'updated_at': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['hipchat']


# Django
from django.contrib.admin.widgets import ForeignKeyRawIdWidget


class GrapevineForeignKeyRawIdWidget(ForeignKeyRawIdWidget):
    def render(self, name, value, attrs=None):
        """
        Extends super.render by adding a Bootstrap class to the ForeignKeyRawId widget
        """
        if attrs is None:
            attrs = {}
        attrs.setdefault('class', '')
        attrs['class'] += ' form-control'

        # Now that we've added the class to attrs, call super.render
        return super(GrapevineForeignKeyRawIdWidget, self).render(name, value, attrs)



from __future__ import unicode_literals
import time
import json

# Django
from django.conf import settings
from django.core.mail.message import EmailMessage
from django.db import transaction

# 3rd Party
# from celery import shared_task
try:
    # This library isn't absolutely required until you
    # want to use the SendGrid backend
    import sendgrid
except ImportError:
    pass


# Local Apps
from .base import GrapevineEmailBackend
from grapevine.settings import grapevine_settings
from grapevine.emails import models as email_models


class EmailBackend(GrapevineEmailBackend):

    DISPLAY_NAME = "sendgrid"
    IMPORT_PATH = "grapevine.emails.backends.SendGridEmailBackend"

    # These are SendGrid variables that have a simple format of
    # `name` - enabled - {0|1}
    SIMPLE_VARIABLE_MAP = {
        "should_show_unsubscribe_link": "subscriptiontrack",
        "should_track_clicks": "clicktrack",
        "should_track_opens": "opentrack"
    }

    # Keys here are SendGrid event names, values are
    # Grapevine event names
    EVENTS_MAP = {
        # Only need to list values here that do not map
        # exactly (after first letter capitalization)
        "spamreport": "Spam Report"
    }

    UNIQUE_ARG_NAME = 'grapevine-guid'

    def __init__(self, fail_silently=False, username=None, password=None, **kwargs):
        # Sets ``self.fail_silently``
        super(EmailBackend, self).__init__(fail_silently=fail_silently)

        try:
            self.username = username or getattr(settings, 'SENDGRID_USERNAME')
            self.password = password or getattr(settings, 'SENDGRID_PASSWORD')
        except AttributeError:
            if fail_silently:
                self.username, self.password = None
            else:
                raise

        self.driver = sendgrid.SendGridClient(self.username, self.password,
            # Don't pass ``not self.fail_silently`` here, as that triggers
            # Exceptions to be raised in the SendGrid code instead of just
            # returning the response, which we would strongly prefer.
            raise_errors=False)

    @classmethod
    def debugify_message(cls, message):
        message.set_tos([grapevine_settings.DEBUG_EMAIL_ADDRESS])
        message.to = [grapevine_settings.DEBUG_EMAIL_ADDRESS]
        message.bcc = []
        return message

    @staticmethod
    def as_message(email):
        """
        Takes a ``grapevine.emails.Email`` model and converts it
        into a ``sendgrid.Mail`` object.
        """
        message = sendgrid.Mail()
        message.add_to(email.to)
        message.add_to(email.cc)
        message.add_bcc(email.bcc)
        message.set_text(email.text_body)
        message.set_html(email.html_body)
        message.set_subject(email.subject)
        message.set_from(email.from_email)
        message.set_replyto(email.reply_to)

        # Grapevine GUID
        message.add_unique_arg(EmailBackend.UNIQUE_ARG_NAME, email.guid)

        for variable in email.variables.all():
            if not bool(variable.value):
                message.add_category(variable.key)
            else:
                if variable.key in EmailBackend.SIMPLE_VARIABLE_MAP.keys():
                    sendgrid_key = EmailBackend.SIMPLE_VARIABLE_MAP[variable.key]
                    message.add_filter(sendgrid_key, 'enabled',
                        int(variable.value))
                elif variable.key[:4] == 'utm_':
                    message.add_filter('ganalytics', 'enabled', 1)
                    message.add_filter('ganalytics', variable.key, variable.value)

        return message

    def from_django_message(self, django_message):
        """
        Converts a ``django.core.mail.messages.EmailMessage`` into a
        ``grapevine.emails.Email`` object, then in turn into a
        ``sendgrid.Mail`` object.
        """
        initial_data = {
            'backend': 'grapevine.emails.backends.SendGridEmailBackend',
            'from_email': django_message.from_email,
            'subject': django_message.subject,
            'text_body': django_message.body,
        }
        if 'reply-to' in django_message.extra_headers:
            initial_data['reply_to'] = django_message.extra_headers.pop('reply-to')

        email = email_models.Email.objects.create(**initial_data)

        email.add_tos(django_message.to)
        email.add_ccs(django_message.cc)
        email.add_bccs(django_message.bcc)

        return self.as_message(email)

    def send_messages(self, email_messages):
        """
        Loops over a list (or single instance) of a Message object
        and sends said message using SendGrid's officially provided
        Python library.
        """
        if not isinstance(email_messages, list):
            email_messages = [email_messages]

        num_sent = 0
        for email_message in email_messages:
            if self.send(email_message):
                num_sent += 1

        return num_sent

    def send(self, email_message):
        """
        Sends a payload to SendGrid. Note that if ``self.fail_silently`` is False
        then the SendGrid library does not bother to return True on a successful
        send, as the lack of a raised Exception signifies a successful send.
        """
        # Cast a raw Django message obj as a Grapevine email. Note that this
        # is only possible if someone uses's Django's off-the-shelf ``send_mail``
        # function with this backend. In that case, the developer sacrifices
        # honoring of their ``UnsubscribedAddress`` list and DEBUG mode logic.
        if isinstance(email_message, EmailMessage):
            email_message = self.from_django_message(email_message)

        # Finally... we pass the email_message object to SendGrid's library
        self.send_response_code, self.send_response_body = self.driver.send(email_message)

        # Arbitrarily return True if ``self.fail_silently``, otherwise
        # only return True if SendGrid agrees that it worked
        return self.fail_silently or self.send_response_code == 200

    def get_event(self, sendgrid_event_name):
        if sendgrid_event_name in self.EVENTS_MAP.keys():
            return self.EVENTS_MAP[sendgrid_event_name]
        else:
            return sendgrid_event_name.capitalize()

    def process_events(self, limit=300):
        self.get_logger().debug("Beginning async_process_events with limit=%s" % limit)
        try:
            # Get the backend obj
            sendgrid_backend = email_models.EmailBackend.objects.filter(path="grapevine.emails.backends.SendGridEmailBackend")[0]
        except IndexError:
            # If there's not even a SendGrid record in the DB then we haven't sent any emails
            # through it and we don't have to worry about processing events
            return None

        with transaction.atomic():
            event_pks = email_models.RawEvent.objects.filter(
                backend=sendgrid_backend,
                processed_on=None,
                is_queued=False).exclude(is_broken=True).values_list('pk', flat=True)[:limit]

            # We must turn this queryset into a raw list of IDs, lest our
            # update to queue the events means eventual outright evaluation
            # of the queryset will re-run the query and return incorrect  records
            event_pks = [event_pk for event_pk in event_pks]

            email_models.RawEvent.objects.filter(pk__in=event_pks).update(is_queued=True)

        for event_pk in event_pks:
            email_models.RawEvent.async_process.delay(event_pk)

    def process_event(self, raw_event):
        """
        Arguments:
        @raw_event  {email_models.RawEvent}  That which we shall process.

        Returns  (bool, float,)   Success flag, time_taken
        """
        start = time.time()

        try:
            # Load the JSON into a dict, return False
            # if we have any problems
            payload = json.loads(raw_event.payload)
        except:
            return False, None

        for raw_event_dict in payload:

            try:
                # We can only process an event containing a
                # known Grapevine guid
                guid = raw_event_dict[self.UNIQUE_ARG_NAME]
            except KeyError:
                continue

            # Figure out which event it is that happened
            event_name = self.get_event(raw_event_dict['event'])
            try:
                event_type = email_models.Event.objects.get(name=event_name)
            except email_models.Event.DoesNotExist:
                continue

            # To which email in our system does this correspond?
            try:
                email = email_models.Email.objects.get(guid=guid)
            except email_models.Email.DoesNotExist:
                # Alarming, but not really anything to do.
                # STOP DELETING RECORDS!
                continue

            # Mint the event record
            email_event, created = email_models.EmailEvent.objects.get_or_create(email=email,
                event=event_type, raw_event=raw_event, happened_at=self.datetime_from_seconds(raw_event_dict['timestamp']))

            # Mark the email address as "Unsubscribed" if it's that kind of Event
            if event_type.should_stop_sending and raw_event_dict.get('email', None):
                email_models.UnsubscribedAddress.objects.create(address=raw_event_dict['email'],
                    email=email)

            # Note the event in the Email's direct log
            email.append_to_log(json.dumps(raw_event_dict))

        return True, (time.time() - start)



from __future__ import unicode_literals
import inspect

# 3rd Party
# from celery import shared_task

# Local Apps
from . import backends


# @shared_task
def async_process_events(limit=300):
    for attr_name in dir(backends):
        attr = getattr(backends, attr_name)
        if inspect.isclass(attr) and issubclass(attr, backends.base.GrapevineEmailBackend) and hasattr(attr, 'process_events'):
            backend = attr()
            backend.process_events(limit)



from __future__ import unicode_literals


def smoosh_args(*args, **kwargs):
    key_suffix = ''
    prepared_args = []
    prepared_kwargs = []
    for arg in args:
        prepared_args.append(str(arg))

    for key, value in kwargs.items():
        prepared_kwargs.append('%s=%s' % (key, str(value),))

    all_prepared_args = prepared_args + prepared_kwargs
    if all_prepared_args:
        key_suffix = ';'.join(all_prepared_args)

    return key_suffix


def memoize(fnc):
    """
    Decorator that accepts a variable name at which to cache the results
    of this function. Now honoring parameters!

    Usage:

        class SomeClass(object):
            @memoize
            def get_title(self):
                print "doing work!"
                return "Jurassic Park"

            @property
            @memoize
            def favorite_trilogy(self):
                print "doing property work!"
                return "Star Wars"


        >>> my_obj = SomeClass()
        >>> my_obj.get_title()
        >>> doing work!
        >>> "Jurassic Park"
        >>>
        >>> my_obj.get_title()
        >>> "Jurassic Park"
        >>>
        >>> my_obj.favorite_trilogy
        >>> doing property work!
        >>> "Star Wars"
        >>>
        >>> my_obj.favorite_trilogy
        >>> "Star Wars"

    """
    def wrapper(self, *args, **kwargs):
        # Make sure the memoize cache
        if not hasattr(self, '_memoize_cache'):
            self._memoize_cache = {}

        cache_attr_name = fnc.__name__
        self._memoize_cache.setdefault(cache_attr_name, {})

        key_suffix = smoosh_args(*args, **kwargs)

        try:
            # Hope for a cache hit
            return self._memoize_cache[cache_attr_name][key_suffix]
        except KeyError:
            self._memoize_cache[cache_attr_name][key_suffix] = None

        # Calculate the value
        val = fnc(self, *args, **kwargs)

        # Save it for later
        self._memoize_cache[cache_attr_name][key_suffix] = val
        return val

    return wrapper



from __future__ import unicode_literals
import uuid
import time
import traceback

# Django
from django.db import models
from django.conf import settings
from django.utils import timezone
# from django.contrib.contenttypes.models import ContentType
try:
    # This is 1.7
    from django.contrib.contenttypes.generic import GenericForeignKey
except ImportError:
    # This is the 1.6 import
    from django.contrib.contenttypes.fields import GenericForeignKey

# 3rd Party
import html2text

# Local Apps
from .models_base import GrapevineModel


class Transport(GrapevineModel):
    """
    The abstract parent class for all various modes of Sendable transportation.
    """
    SENT = 1
    UNSENT = 2
    FAILED = 3
    SEND_TIME_ERROR = 4
    NEWSLETTER = 5
    DUPLICATE = 6
    UNSUBSCRIBED = 7
    STATUS_CHOICES = (
        (SENT, 'Sent',),
        (FAILED, 'Failed',),
        (UNSENT, 'Unsent',),
        (SEND_TIME_ERROR, 'Send-time Error',),
        (NEWSLETTER, 'Newsletter',),
        (DUPLICATE, 'Duplicate',),
        (UNSUBSCRIBED, 'Unsubscribed',),
    )

    type = models.ForeignKey('contenttypes.ContentType', blank=True, null=True, default=None)
    html_body = models.TextField(blank=True)
    text_body = models.TextField(blank=True)
    status = models.IntegerField(choices=STATUS_CHOICES, default=UNSENT, db_index=True)
    sent_at = models.DateTimeField(db_index=True, verbose_name='Sent At', null=True, default=None, blank=True)
    communication_time = models.DecimalField(max_digits=8, decimal_places=5, null=True, blank=True, default=None,
        verbose_name="Communication Time", help_text="In seconds")
    is_test = models.BooleanField(default=False, verbose_name='Is Test', db_index=True)
    guid = models.CharField(max_length=36, null=True, blank=True, db_index=True)
    log = models.TextField(blank=True, null=True, default=None)

    class Meta(GrapevineModel.Meta):
        abstract = True

    def ensure_guid(self):
        if not (self.guid):
            self.guid = str(uuid.uuid4())
        return self

    @property
    def sendable_class(self):
        return self.type.model_class()

    @property
    def sendable(self):
        if not hasattr(self, '_sendable'):
            try:
                self._sendable = self.sendable_class.objects.get(message_id=self.pk)
            except:
                # Purposefully don't cache self._sendable here to allow
                # future code (after this message is sent) to accurately
                # return the ``sendable`` obj
                return None
        return self._sendable

    def determine_text_body(self):
        if not bool(self.text_body):
            self.text_body = html2text.html2text(self.html_body)

    def save(self, *args, **kwargs):
        """
        Enforces some code-heavy (not so much DB-heavy) defaults, then saves.
        """
        # New message records get GUIDs
        self.ensure_guid()

        # Didn't supply a plain text value? You get a converted version
        # from your HTML value for free.
        self.determine_text_body()

        self.pre_save(*args, **kwargs)
        super(Transport, self).save(*args, **kwargs)
        self.post_save(*args, **kwargs)

    def pre_save(self, *args, **kwargs):
        pass

    def post_save(self, *args, **kwargs):
        pass

    def send(self, *args, **kwargs):
        try:
            # Put a stopwatch on the actual communication
            start = time.time()

            is_sent = self._send(*args, **kwargs)

            end = time.time()
            self.communication_time = (end - start)

            # Now that the message has been sent, delete it from
            # the queue to keep lookups to that table nice and fast
            if self.sendable:
                self.sendable.delete_from_queue()

            if is_sent:
                self.status = self.SENT
                self.sent_at = timezone.now()
            else:
                # Honor any status setting that may have happened
                # during ``_send()``
                if self.status == self.UNSENT:
                    self.status = self.FAILED

            self.save()
            return is_sent
        except Exception as e:
            stack = traceback.extract_stack()
            formatted_stack = '\n'.join(traceback.format_list(stack))

            self.status = self.SEND_TIME_ERROR
            self.append_to_log(formatted_stack, should_save=False, desc=e.args[0])
            self.save()
            return False

    def _send(self, *args, **kwargs):
        """
        Actual bits onto the wire fun. Must be implemented by non-abstract
        Transport subclasses.
        """
        raise NotImplementedError("%s class has not implemented ``_send``" % (self.__class__.__name__,))

    @classmethod
    def from_sendable(cls, sendable, recipient_address=None, is_test=False, **kwargs):
        """
        Given an instance of a Sendable, this builds a corresponding instance of
        whichever child ``transport`` class this ultimately is.

        Arguments:
        @sendable           {mixed} An instance of a class using the SendableMixin
        @recipient_address  {str}   A recipient-override option. Useful in conjunction
                                    with `is_test`.
        @is_test            {bool}  A bookkeeping variable to quickly see if this message
                                    was a test.
        """
        initial_data = {
            'type': sendable.content_type(),
            'to': sendable._get_recipients(recipient_address),
            'html_body': sendable.html_body,
            'text_body': sendable.text_body,
            'status': cls.UNSENT,
            'is_test': is_test,
        }
        initial_data = cls.finish_initial_data(sendable, initial_data, **kwargs)
        transport = cls.objects.create(**initial_data)

        # Save the link between sendable and message
        if not is_test:
            sendable.message_id = transport.pk
            sendable.save()

        # Allow the Sendable object to make last second
        # modifications to the Transport. This may include
        # adding tags, altering attributes, etc.
        transport = sendable.alter_transport(transport, **kwargs)

        return transport

    @staticmethod
    def finish_initial_data(sendable, data, **kwargs):
        """
        Hook for using classes to optionally augment the initial data given
        to their Transport object.
        """
        return data


class QueuedMessage(GrapevineModel):
    """
    Records are written to this table when a message is queued, and
    subsequently deleted when the message is ultimately sent.
    """

    message_type = models.ForeignKey('contenttypes.ContentType', verbose_name="Message Type")
    message_id = models.PositiveIntegerField(verbose_name="Message Id")

    message = GenericForeignKey('message_type', 'message_id')

    class Meta:
        verbose_name = "Queued Message"
        verbose_name_plural = "Queued Messages"

    def unicode(self):
        return 'Queued Message %s:%s' % (self.message_type_id, self.message_id,)



from __future__ import unicode_literals


class SynchronousSender(object):

    def send(self, sendable, *args, **kwargs):
        """
        Puts bits directly on the wire.
        """
        # self.get_logger().info("Synchronously sending %s:%s", sendable.__class__.__name__, sendable.pk)
        return sendable.send(*args, **kwargs)


class AsynchronousSender(object):

    def send(self, sendable, *args, **kwargs):
        """
        Pass sendable PK to a queue.
        """
        # self.get_logger().info("Asynchronously sending %s:%s", sendable.__class__.__name__, sendable.pk)
        sendable.denote_as_queued()
        return sendable.async_send.delay(sendable.__class__, sendable.pk, *args, **kwargs)



from __future__ import unicode_literals
import inspect

# Local Apps
from .emails import backends

urlpatterns = []

for attr_name in dir(backends):
    attr = getattr(backends, attr_name)
    if inspect.isclass(attr) and issubclass(attr, backends.base.GrapevineEmailBackend):
        try:
            # TODO: Formally activate these.
            backend = attr()
            urlpatterns += backend.get_urls()
        except:
            pass



from __future__ import unicode_literals

# Django
from django.contrib import admin, messages
from django.http import HttpResponse, HttpResponseRedirect, Http404
from django.conf.urls import url
from django.core.urlresolvers import reverse, NoReverseMatch
from django.shortcuts import get_object_or_404
from django.contrib.contenttypes.models import ContentType
from django.utils import six

# Local Apps
from grapevine.emails.filters import OnSpecificDateListFilter
from grapevine.utils import render_view


class BaseModelAdmin(admin.ModelAdmin):

    additional_object_tool_excludes = ()

    @property
    def model_meta_info(self):
        return (self.model._meta.app_label, self.model._meta.model_name,)

    @property
    def admin_view_info(self):
        return '%s_%s' % self.model_meta_info

    def render_change_form(self, request, context, add=False, change=False, form_url='', obj=None):
        if obj:
            context['additional_object_tools'] = self.additional_object_tools(obj)
        return super(BaseModelAdmin, self).render_change_form(request, context, add, change, form_url, obj)

    def additional_object_tools(self, obj):
        tool_urls = []
        excludes = self.get_additional_object_tool_excludes(obj)
        for relationship in obj._meta.get_all_related_objects():
            # Skip all excludes
            if relationship.get_accessor_name() in excludes:
                continue

            remote_field_name = relationship.field.name
            try:
                url = reverse('admin:%s_%s_changelist' % (relationship.model._meta.app_label, relationship.model._meta.model_name,))
                url += '?%s=%s' % (remote_field_name, obj.pk,)

                display_name = "View %s" % (relationship.get_accessor_name().title(),)
                display_name = display_name.replace('_', ' ')
                tool_urls.append({
                    'url': url,
                    'display_name': display_name
                })
            except NoReverseMatch:
                pass

        return tool_urls

    def get_additional_object_tool_excludes(self, obj):
        """
        Returns an interable of relationship ``get_accessor_name()`` values that should **not** be automatically
        added to the additional tools section in the admin.

        Generally speaking, ``get_accessor_name()`` returns the name of the ReverseManager,
        which is what is overwritten by the ``related_name`` keyword on ForeignKey fields.
        """
        return self.additional_object_tool_excludes


class SendableInline(admin.TabularInline):

    fields = ["admin_id", "admin_message", "scheduled_send_time", "cancelled_at_send_time"]

    def admin_id(self, obj):
        url = reverse('admin:%s_change' % (obj.admin_view_info,), args=(obj.id,))
        return '<a href="%s" target="_blank">%s</a>' % (url, obj.id,)
    admin_id.allow_tags = True
    admin_id.short_description = 'Id'

    def admin_message(self, obj):
        if obj.message is None:
            return '--'

        transport_class_app_name = obj.get_transport_class()._meta.app_label
        transport_class_sluggy_name = obj.get_transport_class()._meta.verbose_name.lower().replace(' ', '_')
        url = reverse('admin:%s_%s_change' % (transport_class_app_name, transport_class_sluggy_name,), args=(obj.message_id,))
        return '<a href="%s" target="_blank" style="text-decoration:none;">%s</a>' % (url, obj.message.__unicode__(),)
    admin_message.allow_tags = True
    admin_message.short_description = 'Message'

    def has_delete_permission(self, request, obj=None):
        return True

    def has_add_permission(self, request, obj=None):
        return True

    readonly_fields = ['admin_id', 'admin_message']
    extra = 0


class SendableAdminMixin(object):
    """
    You must set `model` in any Admin classes inheriting from this
    class or nothing will work.
    """
    preview_height = 100

    # Used for admin display purposes
    message_type_verbose = "Message"

    readonly_fields = ['admin_message']
    list_filter = (('scheduled_send_time', OnSpecificDateListFilter),)

    change_form_template = 'admin/change_sendable_form.html'

    def lookup_allowed(self, key, value):
        """
        Normally list filtering is only *allowed* on things specified in
        ``list_filter`` above, but that's a problem because the Django admin
        wants to draw a dropdown containing all possible options. To search
        on ``model__related_model``, we would have to specify exactly that, and Django
        might try to draw a dropdown of all umpteen million records. No bien.

        Thus, we whitelist some fields ourselves as being legal edits without
        specifying them in the place where a fckng mammoth <select> field could result.
        """
        for field_name in self.whitelisted_filter_fields:
            if field_name in key:
                return True
        return super(SendableAdminMixin, self).lookup_allowed(key, value)

    def admin_message(self, obj):
        if obj.message is None:
            return '--'

        transport_class_app_name = obj.get_transport_class()._meta.app_label
        transport_class_sluggy_name = obj.get_transport_class()._meta.verbose_name.lower().replace(' ', '_')
        url = reverse('admin:%s_%s_change' % (transport_class_app_name, transport_class_sluggy_name,), args=(obj.message_id,))
        return '<a href="%s" target="_blank" style="text-decoration:none;">%s</a>' % (url, obj.message.__unicode__(),)
    admin_message.allow_tags = True
    admin_message.short_description = 'Message'

    def render_change_form(self, request, context, add=False, change=False, form_url='', obj=None):
        if obj:
            # Get the preview URL
            view_name = 'admin:%s_render' % (self.admin_view_info,)
            context['preview_height'] = self.preview_height
            context['preview_url'] = reverse(view_name, args=(obj.pk,))

            # Add in the message type, for nice clarity across various types of transports
            context['message_type_verbose'] = self.message_type_verbose
        return super(SendableAdminMixin, self).render_change_form(request, context, add, change, form_url, obj)

    def additional_object_tools(self, obj):
        tools = super(SendableAdminMixin, self).additional_object_tools(obj)
        tools.append({
            'url': reverse('admin:%s_send_test_message' % (self.admin_view_info,), args=(obj.pk,)),
            'display_name': 'Send Test %s' % (self.message_type_verbose,)
        })
        if not obj.is_sent or not obj.message or obj.message.status != obj.message.SENT:
            tools.append({
                'url': reverse('admin:%s_send_real_message' % (self.admin_view_info,), args=(obj.pk,)),
                'display_name': 'Send Real %s' % (self.message_type_verbose,)
            })
        return tools

    @property
    def model_meta_info(self):
        return (self.model._meta.app_label, self.model._meta.model_name,)

    @property
    def admin_view_info(self):
        return '%s_%s' % self.model_meta_info

    def get_urls(self):
        urls = super(SendableAdminMixin, self).get_urls()
        my_urls = [
            url(r'^(.+)/render/$', self.admin_site.admin_view(self.render), name='%s_render' % (self.admin_view_info,)),
            url(r'^(.+)/send-test/$', self.admin_site.admin_view(self.send_test_message), name='%s_send_test_message' % (self.admin_view_info,)),
            url(r'^(.+)/send-real/$', self.admin_site.admin_view(self.send_real_message), name='%s_send_real_message' % (self.admin_view_info,)),
        ]
        return my_urls + urls

    def send_real_message(self, request, obj_id):
        obj = get_object_or_404(self.model, pk=obj_id)
        if request.method == 'GET':
            recipients = obj.get_recipients()
            if isinstance(recipients, six.string_types):
                recipients = [recipients]
            context = {
                'obj': obj,
                'recipients': recipients,
                'is_recipients_dict': isinstance(recipients, dict),
                'opts': self.model._meta,
                'title': 'Send Real %s' % (self.message_type_verbose,)
            }
            return render_view(request, 'admin/send_real.html', context)
        elif request.method == 'POST':
            return self.send_message(request, obj, False)

    def send_test_message(self, request, obj_id):
        obj = get_object_or_404(self.model, pk=obj_id)
        if request.method == 'GET':
            context = {
                'recipients': self.get_test_recipient(request, obj_id),
                'opts': self.model._meta,
                'title': 'Send Test %s' % (self.message_type_verbose,)
            }
            return render_view(request, 'admin/send_test.html', context)
        elif request.method == 'POST':
            return self.send_message(request, obj, True, request.POST.get('recipient_address'))

    def get_test_recipient(self, request, obj_id):
        """
        No idea what makes sense against a generic Sendable
        """
        return ''

    def send_message(self, request, obj, is_test, recipient_address=None):
        if not request.user.is_authenticated() or not request.user.is_staff:
            raise Http404

        if is_test and not request.user.email:
            messages.add_message(request, messages.ERROR,
                "Current Admin user does not have a specified email address.")
            return HttpResponseRedirect(reverse("admin:%s_change" % (self.admin_view_info,), args=(obj.pk,)))

        # Load the Sendable and send a message
        is_sent = obj.send(recipient_address=recipient_address, is_test=is_test, force_resend=True)

        if is_sent:
            message_beginning = 'Test' if is_test else ''
            message_ending = 'to %s' % (recipient_address) if recipient_address else ''
            messages.add_message(request, messages.SUCCESS,
                "%s Message sent %s" % (message_beginning, message_ending,))

            content_type_id = ContentType.objects.get_for_model(obj).pk
            admin.models.LogEntry.objects.log_action(
                user_id=request.user.pk,
                content_type_id=content_type_id,
                object_id=obj.pk,
                object_repr=obj.__unicode__(),
                action_flag=admin.models.CHANGE,
                change_message='Sent %s Message Id: %s for %s Id: %s' % (message_beginning, obj.transport.pk, obj._meta.verbose_name, obj.pk,)
            )
        else:
            messages.add_message(request, messages.ERROR,
                "Problem sending email.")

        return HttpResponseRedirect(reverse('admin:%s_change' % (self.admin_view_info,), args=(obj.pk,)))

    def render(self, request, obj_id):
        if not request.user.is_authenticated() or not request.user.is_staff:
            raise Http404

        obj = get_object_or_404(self.model, pk=obj_id)
        try:
            return HttpResponse(obj.render())
        except Exception as e:
            return HttpResponse("%s: %s" % (e.__class__.__name__, e.args[0],))



from __future__ import unicode_literals

import requests

# Django
from django.db import models
from django.conf import settings
from django.utils import timezone, six, module_loading
from django.utils.html import strip_tags

# 3rd Party
import html2text

# Local Apps
from grapevine.models import BaseModel
from grapevine.settings import grapevine_settings
from grapevine.models import Transport
from core import utils
from core.utils import logging_class

@logging_class
class HipChat(Transport):
    GENERAL_TALKS = 130723
    DATABASE_DISCUSSIONS = 130743
    GITHUB = 150746
    SOLUTIONS_TEAM = 202172
    JENKINS = 208896
    QA = 230576
    GROWTH_HACKING = 345591
    EMAILS = 362339
    BACKEND_TEAM = 391362
    NN_DESIGNS = 464123
    REVIEW_TASKFORCE = 470669
    SERVER_ALERTS = 473245
    ADS_TEAM = 473779
    FRONT_END_DISCUSSIONS = 474520
    CAMPAIGN_TEAM = 494179
    IVORY = 497734
    CBWM = 498360
    CASE_STUDY = 546354
    SALESFORCE = 563824
    SUPPORT = 581333
    ROOM_CHOICES = (
        (GENERAL_TALKS, 'General Talks'),
        (DATABASE_DISCUSSIONS, 'Database Discussions'),
        (GITHUB, 'Github'),
        (SOLUTIONS_TEAM, 'Solutions Team'),
        (JENKINS, 'Jenkins'),
        (QA, 'QA'),
        (GROWTH_HACKING, 'Growth Hacking'),
        (EMAILS, 'Emails'),
        (BACKEND_TEAM, 'Backend Team'),
        (NN_DESIGNS, '99 designs'),
        (REVIEW_TASKFORCE, 'Review Taskforce'),
        (SERVER_ALERTS, 'Server Alerts'),
        (ADS_TEAM, 'Ads team'),
        (FRONT_END_DISCUSSIONS, 'Front End Discussions'),
        (CAMPAIGN_TEAM, 'Campaign Team'),
        (IVORY, 'Ivory'),
        (CBWM, 'CBWM'),
        (CASE_STUDY, 'Case Study'),
        (SALESFORCE, 'Salesforce'),
        (SUPPORT, 'Support'),
    )
    to = models.IntegerField(choices=ROOM_CHOICES, default=IVORY, help_text="The HipChat Room Id")


    YELLOW = 'yellow'
    RED = 'red'
    GREEN = 'green'
    PURPLE = 'purple'
    GRAY = 'gray'
    COLOR_CHOICES = (
        (YELLOW, 'Yellow'),
        (RED, 'Red'),
        (GREEN, 'Green'),
        (PURPLE, 'Purple'),
        (GRAY, 'Gray'),
    )
    color = models.CharField(max_length=50, choices=COLOR_CHOICES, default=YELLOW)

    FORMAT_HTML = 'html'
    FORMAT_TEXT = 'gray'
    FORMAT_CHOICES = (
        (FORMAT_HTML, 'HTML'),
        (FORMAT_TEXT, 'Text'),
    )
    message_format = models.CharField(max_length=50, choices=FORMAT_CHOICES, default=FORMAT_HTML)


    NO = 0
    YES = 1
    NOTIFY_CHOICES = (
        (YES, 'Yes'),
        (NO, 'No'),
    )
    should_notify = models.IntegerField(choices=NOTIFY_CHOICES, default=NO)

    from_name = models.CharField(max_length=255)

    class Meta:
        verbose_name = 'HipChat'
        verbose_name_plural = 'HipChats'

    @staticmethod
    def finish_initial_data(sendable, initial_data, **kwargs):
        initial_data['to'] = sendable.TO
        initial_data['color'] = sendable.MESSAGE_COLOR
        initial_data['should_notify'] = sendable.SHOULD_NOTIFY
        initial_data['from_name'] = sendable.FROM_NAME
        return initial_data

    @property
    def message(self):
        if self.message_format and self.message_format == self.FORMAT_HTML:
            return self.html_body
        else:
            return self.text_body

    @message.setter
    def message(self, value):
        """
        Set both values logically
        """
        self.html_body = value
        self.text_body = strip_tags(value)

    @property
    def hipchat_url(self):
        return "https://api.hipchat.com/v1/rooms/message?auth_token=%s" % settings.HIPCHAT_TOKEN

    def _send(self):
        """
        Required. Send the hipchat notification
        """
        logger = utils.create_logger(HipChat, '_send')

        try:
            if self.message and self.from_name:
                self.resp = requests.post(
                    self.hipchat_url,
                    params={
                        "room_id": self.to,
                        "from": self.from_name,
                        "message": self.message,
                        "color": self.color,
                        "format": self.message_format,
                        "notify": self.should_notify
                    })

                if self.resp.status_code == 200:
                    return True
                else:
                    logger.error('HipChat Error: %s' % e)
        except Exception as e:
            logger.error(e)

        return False



"""
Grapevine settings are namespaced in the GRAPEVINE setting.
For example, your project's `settings.py` file might look like this:

GRAPEVINE = {
    'EMAIL_BACKEND' = 'grapevine.emails.backends.SendGridEmailBackend',
}

**Nod to Django Rest Framework's structured settings setup.**
"""
from __future__ import unicode_literals

# Django
from django.conf import settings
from django.utils import importlib, six


USER_SETTINGS = getattr(settings, 'GRAPEVINE', {})
DEFAULTS = {
    'SENDER_CLASS': 'grapevine.engines.SynchronousSender',
    'EMAIL_BACKEND': None,
    'DEBUG_EMAIL_ADDRESS': 'test@email.com',
}

# These values, if unspecified, fallback to their
# global settings values
GLOBAL_DEFAULTS = (
    'EMAIL_BACKEND',
    'DEBUG',
)

# List of settings that will be dotted import paths
IMPORTABLE_SETTINGS = (
    'SENDER_CLASS',
)


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if isinstance(val, six.string_types):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        # Nod to tastypie's use of importlib.
        parts = val.split('.')
        module_path, class_name = '.'.join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except ImportError as e:
        msg = "Could not import '%s' for API setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


class GrapevineSettings(object):
    """
    A settings object that allows Grapevine to access user-specified settings
    values, with appropriate default fallbacks.

        from grapevine.settings import grapevine_settings
        if grapevine_settings.DEBUG:
            # Do stuff
        else:
            # Do other stuff
    """

    def __init__(self, user_settings=None, defaults=None,
                 global_settings_keys=None, importable_settings=None, django_settings=None):
        self.user_settings = user_settings or {}
        self.defaults = defaults or {}
        self.global_settings_keys = global_settings_keys or {}
        self.importable_settings = importable_settings or ()
        self.django_settings = django_settings or settings

    def __getattr__(self, attr):
        if attr not in self.defaults.keys() and attr not in self.global_settings_keys:
            raise AttributeError("Invalid Grapevine setting: '%s'" % (attr,))

        try:
            # First try user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults.get(attr, None)

        # Fallback to global settings if appropriate
        if val is None and attr in self.global_settings_keys:
            val = getattr(settings, attr, None)

        # Coerce import strings into classes
        if val and attr in self.importable_settings:
            val = perform_import(val, attr)

        # Cache the result
        setattr(self, attr, val)
        return val

grapevine_settings = GrapevineSettings(USER_SETTINGS, DEFAULTS, GLOBAL_DEFAULTS, IMPORTABLE_SETTINGS)



from __future__ import unicode_literals

# Local Apps
from . import models
from .jobs import async_process_events


default_app_config = 'grapevine.emails.apps.EmailsConfig'


def send_mail(subject, message, from_email, recipient_list, html_message='',
            fail_silently=False, auth_user=None, auth_password=None,
            connection=None):
    """
    Overrides Django's native ``send_mail`` function to add standard Grapevine
    functionality like unsubscribe honoring and DEBUG mode sending.
    """

    email = models.Email.objects.create(
        subject=subject,
        html_body=html_message,
        text_body=message,
        from_email=from_email,
        to=recipient_list,
    )

    return email.send()



from __future__ import unicode_literals

# Django
from django.db import models

# Local Apps
from grapevine.emails.utils import parse_email


class EmailManager(models.Manager):

    def create(self, **kwargs):
        """
        Takes and appropriately processes the extra keyword args:
            `recipients` and `{backend|backend_id}`
        """
        recipients = kwargs.pop('to', None)

        key, backend = self.model.determine_backend(**kwargs)
        if backend:
            # If we matched a backend, great
            kwargs[key] = backend

        obj = super(EmailManager, self).create(**kwargs)

        if recipients:
            obj.add_recipients(recipients)

        return obj


class EmailRecipientManager(models.Manager):

    def create(self, **kwargs):
        """
        Handles normal `create` duties, plus some additional "address" parsing
        """
        address = kwargs.get('address', '')
        name, address = parse_email(address)
        kwargs["name"] = name
        kwargs["address"] = address

        return super(EmailRecipientManager, self).create(**kwargs)



from __future__ import unicode_literals
import json

# Django
from django.db import models, transaction
from django.core.urlresolvers import reverse
from django.contrib.contenttypes.models import ContentType
from django.utils import timezone


class GrapevineModel(models.Model):

    # Bookkeeping
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

    @property
    def meta_info(self):
        return (self._meta.app_label, self._meta.model_name,)

    @property
    def admin_view_info(self):
        return '%s_%s' % self.meta_info

    @property
    def admin_uri(self):
        return reverse('admin:%s_%s_change' % (self._meta.app_label, self._meta.model_name,), args=(self.pk,))

    @classmethod
    def content_type(cls):
        return ContentType.objects.get_for_model(cls)

    @classmethod
    def field_names(cls):
        """
        Returns the names of all fields on the model
        """
        return [field.name for field in cls._meta.fields]

    def finalize_serialization(self, serialized, strip_empty=False):
        """
        A hook for child classes to do more phun stuff.
        """
        if strip_empty:
            for key, value in serialized.items():
                if value is None or (isinstance(value, dict) and value['id'] is None):
                    serialized.pop(key)
        return serialized

    def serialize(self, full=False, id_nested=False, strip_empty=False, excludes=[]):
        """
        Returns a dictionary of all local fields.
        """
        serialized = {field.name: self.get_field_value(field.name, full) for field in self.__class__._meta.fields if field.name not in excludes}
        return self.finalize_serialization(serialized, strip_empty=strip_empty)

    def get_field_value(self, field_name, full=False):
        """
        Returns a given value of for this instantiated model.

        Arguments:
        field_name    {string}      The value of the attr you want
        full          {bool}        OPTIONAL. If the passed name is a relation, should it be hydrated?
                                    Defaults to False.
        """
        field = self._meta.get_field(field_name)
        # Is this a related field or a literal?
        if isinstance(field, models.fields.related.RelatedField):
            if full:
                # It's related and they ordered it hydrated
                val = getattr(self, field_name, None)
                # Pull out the value and hydrate it if it exists, else
                # return None
                if val is not None:
                    return val.field_values()  # Don't forward `full` to avoid cyclical problems
                else:
                    return None
            else:
                # Not hydrated is easy enough, just return the PK we
                # already have on hand
                _id = getattr(self, '%s_id' % (field_name,), None)
                return {'id': _id}
        elif isinstance(field, models.fields.DateField):  # Covers both DateTimeField and DateField
            return self._meta.get_field(field_name).value_to_string(self)
        else:
            # Not related? Too easy.
            return getattr(self, field_name, None)

    def reload(self):
        """
        In place DB update of the record.
        """
        new_self = self.__class__.objects.get(pk=self.pk)
        self.__dict__.update(new_self.__dict__)
        return self

    def append_to_log(self, log, should_save=True, desc=None, should_use_transaction=True,
                      should_reload=True):
        """
        The wrapper function around ``_append_to_log()``. The distinction exists to help
        sanely enforce the ``should_use_transaction`` flag.
        """
        if should_save and should_use_transaction:
            with transaction.atomic():
                return self._append_to_log(log, should_save, desc, should_reload)
        else:
            return self._append_to_log(log, should_save, desc, should_reload)

    def _append_to_log(self, log, should_save=True, desc=None, should_reload=True):
        """
        Does the actual work of formatting a log, optional name, and timestamp into
        a block of text that gets appended to the ``self.log``.
        """
        assert 'log' in self.field_names(), "Cannot call ``append_to_log()`` on \
            model without a field named ``log``."

        if self.log is None:
            self.log = ''
            self.save()

        if isinstance(log, dict):
            log = json.dumps(log)

        if should_reload:
            # Update local data to ensure sure nothing else committed something
            # to this record before we entered this transaction.
            self.reload()

        pre_log = '##################\n'
        pre_log += '%s\n' % timezone.now().strftime('%b %d, %Y, %I:%M:%S %p UTC')
        if desc:
            pre_log += '**%s**\n' % desc

        post_log = '\n##################\n'

        self.log += '%s%s%s' % (pre_log, log, post_log,)
        if should_save:
            self.save()

        return self



from __future__ import unicode_literals

# Django
from django.contrib.contenttypes.models import ContentType

# 3rd Party
# from celery import shared_task

# Local Apps
from grapevine.settings import grapevine_settings


class ScheduledSendableSender(object):
    """
    Loops over all ContentTypes and sends each eligible message
    that extend SendableMixin.
    """
    def __init__(self):
        pass

    @staticmethod
    # @shared_task(queue="sendable_queuer")
    def async_deliver_messages():
        sender = ScheduledSendableSender()
        return sender.deliver_messages()

    def deliver_messages(self):
        from grapevine import mixins

        # self.get_logger().info("Beginning scan for sendable messages...")

        num_sent = num_sendables = 0
        content_types = [ct for ct in ContentType.objects.all()]
        for content_type in content_types:
            cls = content_type.model_class()

            # Stale content types will return ``None``
            # above, so make sure to skip such results
            if not cls:
                continue

            if issubclass(cls, mixins.SendableMixin):
                num_sendables += 1
                num_sent += ScheduledSendableSender.process_sendable_model(cls)

        # self.get_logger().info("Sent %s messages for %s total Sendable models.", num_sent, num_sendables)

        return num_sent, num_sendables

    @staticmethod
    def process_sendable_model(cls):
        """
        Arguments:
        cls     {mixed}     A non-abstract Django model that extends one of
                            Grapevine's SendableMixin classes.
        """
        # Create a single instance of the class that actually pushes
        # the Sendable's ``send`` button, or queues up that pushing
        sender = grapevine_settings.SENDER_CLASS()

        # Loop over all unsent, but eligible
        num_sent = 0
        sendable_objs = [obj for obj in ScheduledSendableSender.all_eligible_by_sendable(cls)]
        for sendable_obj in sendable_objs:
            # Some objects may have been scheduled for a long time and
            # require an individual check before sending that would be
            # too expensive to place into ``all_eligible()``.
            should_send = sendable_obj.confirm_individual_sendability()
            if should_send:
                is_sent = sender.send(sendable_obj)
                if is_sent:
                    num_sent += 1
        return num_sent

    @staticmethod
    def all_eligible_by_sendable(cls):
        """
        Component of `process_sendable` broken out for testing purposes.
        """
        return cls.objects.is_eligible()



from __future__ import unicode_literals


def parse_email(address):
    """
    Returns "Marco Polo", "marco@polo.com" from "Marco Polo <marco@polo.com"
    Returns "", "marco@polo.com" from "marco@polo.com"
    """
    if ' ' in address.strip():
        assert '<' in address and '>' in address, "Invalid address structure: %s" % (address,)
        parts = address.split(' ')
        name = ' '.join(parts[:-1])
        address = parts[-1][1:-1]

        return name, address
    else:
        return "", address


class EventRepo(object):
    _instance = None
    name_map = {}

    def __new__(cls, *args, **kwargs):
        """
        Singleton implementation
        """
        if not cls._instance:
            cls._instance = super(EventRepo, cls).__new__(cls, *args, **kwargs)
            cls._instance.seed_name_map()
        return cls._instance

    def seed_name_map(self):
        from models import Event
        self.name_map = {}
        for ev in Event.objects.all():
            self._instance.name_map[ev.name] = ev

    def get_event_by_name(self, name):
        return self.name_map[name]

    def __getitem__(self, name):
        return self.get_event_by_name(name)



from __future__ import unicode_literals

# Django
from django.db import models
from django.conf import settings
from django.utils import timezone
from django.template import Context
from django.template.loader import get_template
try:
    # This is the 1.7 import
    from django.contrib.contenttypes.fields import GenericRelation
except ImportError:
    # And this is the 1.6 import
    from django.contrib.contenttypes.generic import GenericRelation

# 3rd Party
import html2text
from tablets.models import Template
# from celery import shared_task

# Local Apps
from . import emails
from . import models as gv_models
from .managers import SendableManager
from .querysets import SendableQuerySet
from grapevine.decorators import memoize
from grapevine.utils import simple_render


class SendableMixin(models.Model):
    """
    Put this on models in other apps that are "sendable"

    Usage:
        # ``initial_data`` is whatever you need for your model
        sendable = SomeSendableClass.objects.create(**initial_data)
        sendable.send()
    """
    scheduled_send_time = models.DateTimeField(blank=True, default=None, null=True,
        db_index=True, verbose_name="Scheduled Send Time")
    cancelled_at_send_time = models.BooleanField(default=False, verbose_name="Cancelled at Send Time")
    message_id = models.PositiveIntegerField(db_index=True, blank=True, null=True,
        default=None, verbose_name="Message Id")

    objects = SendableManager.for_queryset_class(SendableQuerySet)()

    class Meta:
        abstract = True
        index_together = (
            ('scheduled_send_time', 'cancelled_at_send_time', 'message_id',),
        )

    def save(self, *args, **kwargs):
        # When creating, default ``scheduled_send_time`` to the callable
        if self.scheduled_send_time is None and not self.pk:
            self.scheduled_send_time = self.default_scheduled_send_time()
        return super(SendableMixin, self).save(*args, **kwargs)

    @property
    def is_sent(self):
        return bool(self.message_id)

    @property
    def is_spam(self):
        return self.is_sent and self.message.is_spam

    @property
    def is_opened(self):
        return self.is_sent and self.message.is_opened

    @property
    def is_clicked(self):
        return self.is_sent and self.message.is_clicked

    @property
    @memoize
    def message(self):
        """
        A helper property to ease getting of the associated message.
        """
        if hasattr(self, "_message"):
            return self._message

        if not self.is_sent:
            return None

        try:
            self._message = self.get_transport_class().objects.get(pk=self.message_id)
            return self._message
        except self.get_transport_class().DoesNotExist:
            # This should never happen!
            return None

    @message.setter
    def message(self, value):
        """
        A helper property to ease setting of the associated message.
        """
        self._message = value
        self.message_id = value.pk

    def denote_as_queued(self):
        """
        Once a message is in the queue, we store it in QueuedMessage to avoid
        the sender process continually re-queuing it until actual delivery.
        """
        queued_message, created = gv_models.QueuedMessage.objects.\
            get_or_create(message_type=self.content_type(), message_id=self.pk)
        return queued_message

    def delete_from_queue(self):
        """
        Once a message has passed through the queue and been delivered,
        we delete its QueuedMessage record. Note that this is *not* actually
        deleting the record from the real SQS queue.
        """
        gv_models.QueuedMessage.objects.\
            filter(message_type=self.content_type(), message_id=self.pk).delete()

    def default_scheduled_send_time(self):
        return timezone.now()

    def get_template(self, template_name=None):
        """
        Assumes a traditional Django, file system-based template loader
        """
        if template_name is None:
            template_name = self.get_template_name()
        return get_template(template_name)

    def get_template_name(self):
        raise NotImplementedError("%s failed to implemented `get_template_name`" % self.__class__.__name__)

    def compile_context(self):
        """
        This is the function classes using the SendableMixin should
        override to correctly populate their template.
        """
        return {
            'sendable': self
        }

    def get_context(self):
        """
        This function is used by Transport classes to render a template.
        As such, it should not need to be modified in any way by child classes.
        """
        if not hasattr(self, '_context'):
            self._context = self.compile_context()
        return self._context

    def _get_recipients(self, recipient_address=None):
        # Pass this in when sending test emails to override who
        # would normally receive this sendable.
        if recipient_address is not None:
            return recipient_address

        # If this is a real email and no explicit recipients have
        # been defined, fall back to what this Sendable defines.
        return self.get_recipients()

    def get_recipients(self):
        raise NotImplementedError("%s failed to implement `get_recipients`" % self.__class__.__name__)

    def get_raw_subject(self):
        """
        This function should return the raw subject with possible {{placeholder}}
        values still intact. They will be populated during `get_subject()` against
        self.context()
        """
        raise NotImplementedError("%s failed to implement `get_raw_subject`" % self.__class__.__name__)

    def get_subject(self):
        """
        Accepts the raw subject with possible placeholders and populates them
        against ``self.get_context()``.
        """
        # Get the raw subject
        subject = self.get_raw_subject()

        # Render and return
        return simple_render(subject, self.get_context())

    def get_raw_from_email(self):
        return settings.DEFAULT_FROM_EMAIL

    def get_from_email(self):
        from_email = self.get_raw_from_email()
        return simple_render(from_email, self.get_context())

    def get_raw_reply_to(self):
        return settings.DEFAULT_REPLY_TO

    def get_reply_to(self):
        reply_to = self.get_raw_reply_to()
        return simple_render(reply_to, self.get_context())

    @property
    def text_body(self):
        if not hasattr(self, '_text_body'):
            self._text_body = html2text.HTML2Text().handle(self.html_body)
        return self._text_body

    @property
    def html_body(self):
        """
        To avoid repeated rendering work, we cache the finalized value and
        access the html body through said cache whenever possible. Calling
        `render` outside of here should be a rare occurance.
        """
        if not hasattr(self, '_html_body'):
            self._html_body = self.render()
        return self._html_body

    @html_body.setter
    def html_body(self, value):
        self._html_body = value

    def render(self, template_name=None, context=None):
        """
        Gets ye some HTML
        """
        if context is None:
            context = self.get_context()

        template = self.get_template(template_name)
        return self._render(template, context)

    def _render(self, template, context):
        """
        Override hook for child classes.
        """
        return template.render(Context(context))

    def alter_transport(self, transport, **kwargs):
        """
        Hook for child classes to execute last-second operations on
        the Transport object.
        """
        return transport

    def get_transport_class(self):
        """
        Combine with transport-specific mixins to get this functionality.
        """
        raise NotImplementedError("%s class failed to implement ``get_transport_class``" % (self.__name__.__class__,))

    def as_transport(self, recipient_address=None, is_test=False, **kwargs):
        return self.get_transport_class().from_sendable(self, recipient_address=recipient_address,
            is_test=is_test, **kwargs)

    @staticmethod
    # @shared_task(queue="sendables")
    def async_send(cls, sendable_id, *args, **kwargs):
        sendable = cls.objects.get(id=sendable_id)
        return sendable.send(*args, **kwargs)

    def send(self, recipient_address=None, is_test=False, **kwargs):
        """
        Gets bits on the wire.

        Arguments
        @recipient_address  {str}   Override the recipient by supplying this value. Good for sending test emails.
        @is_test            {bool}  Set to True when overriding the recipient. A bookkeeping variable.
        """
        # Already sent messages shouldn't be resent.
        if not is_test and self.is_sent and not kwargs.pop('force_resend', False):
            # self.get_logger(decorated_class=SendableMixin).warning("Attempted to resend %s with Id %s", self.__class__.__name__, self.pk,)
            return False

        self.transport = self.as_transport(recipient_address=recipient_address,
            is_test=is_test, **kwargs)
        return self.transport.send()

    def send_test(self, recipient_address, **kwargs):
        kwargs['is_test'] = True
        return self.send(recipient_address=recipient_address, **kwargs)

    def confirm_individual_sendability(self):
        """
        Invoked by the actual sender job at the moment of sending.
        Extending classes can override this function and are encouraged,
        if they return False, to also push back ``self.scheduled_send_time``.

        Returns True to send, False to delay
        """
        return True


class TemplateSendableMixin(SendableMixin):
    """
    Exactly the same as the SendableMixin, except for how the message's
    actual content is determined.
    """
    template = models.ForeignKey('tablets.Template', blank=True, default=None, null=True)

    class Meta(SendableMixin.Meta):
        abstract = True

    def get_template_name(self):
        return self.template.name

    def get_raw_subject(self):
        if self.template.subject:
            return self.template.subject
        else:
            return super(TemplateSendableMixin, self).get_raw_subject()

    def get_raw_reply_to(self):
        reply_to = None

        # First ask the template for ``reply_to``
        if self.template:
            reply_to = self.template.reply_to

        # If the Template opted out, ask the model instance
        if not reply_to:
            reply_to = super(TemplateSendableMixin, self).get_raw_reply_to()

        # Lastly, fall back to the settings
        if not reply_to:
            reply_to = settings.DEFAULT_REPLY_TO

        return reply_to

    def get_raw_from_email(self):
        from_email = None

        # First ask the template for ``from_email``
        if self.template:
            from_email = self.template.from_email

        # If the Template opted out, ask the model instance
        if not from_email:
            from_email = super(TemplateSendableMixin, self).get_raw_from_email()

        # Lastly, fall back to the settings
        if not from_email:
            from_email = settings.DEFAULT_FROM_EMAIL

        return from_email

    def get_template(self, template_name=None):
        """
        For duck-typing purposes, this function must accept the ``template_name``
        param, though it makes no sense to do anything with it. We have a hard
        foreign key reference to a template record, and we will use that.
        """
        return self.template.as_template()


class Emailable(object):
    """
    Combine with any of the Sendable mixins to explicitly define which
    Transport class should be used.
    """

    @staticmethod
    def get_transport_class():
        return emails.models.Email

    def remove_unsubscribe_link(self, transport):
        transport.add_variable('should_show_unsubscribe_link', 0)
        return transport

    def track_clicks(self, transport):
        """
        Clicks appear to be tracked automatically.
        """
        transport.add_variable('should_track_clicks', 1)
        return transport

    def track_opens(self, transport):
        """
        Opens appear to be tracked automatically.
        """
        transport.add_variable('should_track_opens', 1)

    def add_ganalytic(self, transport, key, value):
        transport.add_variable('utm_%s' % (key,), value)
        return transport



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='QueuedMessage',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('message_id', models.PositiveIntegerField(verbose_name='Message Id')),
                ('message_type', models.ForeignKey(verbose_name='Message Type', to='contenttypes.ContentType')),
            ],
            options={
                'verbose_name': 'Queued Message',
                'verbose_name_plural': 'Queued Messages',
            },
            bases=(models.Model,),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Email',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('html_body', models.TextField(blank=True)),
                ('text_body', models.TextField(blank=True)),
                ('status', models.IntegerField(default=2, db_index=True, choices=[(1, 'Sent'), (3, 'Failed'), (2, 'Unsent'), (4, 'Send-time Error'), (5, 'Newsletter'), (6, 'Duplicate'), (7, 'Unsubscribed')])),
                ('sent_at', models.DateTimeField(default=None, null=True, verbose_name='Sent At', db_index=True, blank=True)),
                ('communication_time', models.DecimalField(decimal_places=5, default=None, max_digits=8, blank=True, help_text='In seconds', null=True, verbose_name='Communication Time')),
                ('is_test', models.BooleanField(default=False, db_index=True, verbose_name='Is Test')),
                ('guid', models.CharField(db_index=True, max_length=36, null=True, blank=True)),
                ('log', models.TextField(default=None, null=True, blank=True)),
                ('from_email', models.CharField(default='Craig Labenz <admins@djangograpevine.com>', max_length=255, db_index=True)),
                ('reply_to', models.EmailField(default='help@djangograpevine.com', max_length=255, blank=True)),
                ('subject', models.CharField(default='Hello from Grapevine!', max_length=255)),
            ],
            options={
                'verbose_name': 'Email',
                'verbose_name_plural': 'Emails',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='EmailBackend',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('path', models.CharField(help_text='The dotted import path to this backend, in a structure that would satisfy settings.EMAIL_BACKEND.', max_length=255)),
                ('username', models.CharField(help_text='Depending on the provider, this may actually be an API key.', max_length=255, blank=True)),
                ('password', models.CharField(help_text='Depending on the provider, this may actually be an API secret.', max_length=255, blank=True)),
            ],
            options={
                'verbose_name': 'Email Backend',
                'verbose_name_plural': 'Email Backends',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='EmailEvent',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('happened_at', models.DateTimeField(verbose_name='Happened At', db_index=True)),
                ('email', models.ForeignKey(related_name='events', to='emails.Email')),
            ],
            options={
                'verbose_name': 'Email Event',
                'verbose_name_plural': 'Email Events',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='EmailRecipient',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('address', models.EmailField(max_length=254, verbose_name='Recipient Address', db_index=True)),
                ('domain', models.CharField(max_length=100, db_index=True)),
                ('name', models.CharField(max_length=150, blank=True)),
                ('type', models.IntegerField(default=1, choices=[(1, 'To'), (2, 'CC'), (3, 'BCC')])),
                ('email', models.ForeignKey(related_name='recipients', to='emails.Email')),
            ],
            options={
                'verbose_name': 'Email Recipient',
                'verbose_name_plural': 'Email Recipients',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='EmailVariable',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('key', models.CharField(max_length=100)),
                ('value', models.CharField(help_text='1 represents True, and 0 represents False.', max_length=5000, blank=True)),
                ('email', models.ForeignKey(related_name='variables', to='emails.Email')),
            ],
            options={
                'verbose_name': 'Email Variable',
                'verbose_name_plural': 'Email Variables',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=100)),
                ('should_stop_sending', models.BooleanField(default=False, verbose_name='Should Stop Sending')),
            ],
            options={
                'verbose_name': 'Event',
                'verbose_name_plural': 'Events',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='RawEvent',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('payload', models.TextField()),
                ('processed_on', models.DateTimeField(default=None, null=True, verbose_name='Processed On', blank=True)),
                ('processed_in', models.DecimalField(decimal_places=4, default=None, max_digits=5, blank=True, null=True, verbose_name='Processed In')),
                ('is_queued', models.BooleanField(default=False, db_index=True, verbose_name='Is Queued')),
                ('is_broken', models.NullBooleanField(default=None, verbose_name='Is Broken', db_index=True)),
                ('remote_ip', models.GenericIPAddressField(verbose_name='Remote IP', db_index=True)),
                ('backend', models.ForeignKey(to='emails.EmailBackend')),
            ],
            options={
                'verbose_name': 'Raw Event',
                'verbose_name_plural': 'Raw Events',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='UnsubscribedAddress',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('address', models.CharField(max_length=255, db_index=True)),
                ('email', models.ForeignKey(related_name='unsubscribed_addresses', default=None, blank=True, to='emails.Email', help_text='Optional link back to the Email in which this recipient address         clicked Unsubscribe.', null=True)),
            ],
            options={
                'verbose_name': 'Unsubscribed Address',
                'verbose_name_plural': 'Unsubscribed Addresses',
            },
            bases=(models.Model,),
        ),
        migrations.AlterIndexTogether(
            name='emailvariable',
            index_together=set([('key', 'value')]),
        ),
        migrations.AddField(
            model_name='emailevent',
            name='event',
            field=models.ForeignKey(related_name='email_events', to='emails.Event'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='emailevent',
            name='raw_event',
            field=models.ForeignKey(related_name='email_events', verbose_name='Raw Event', to='emails.RawEvent'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='email',
            name='backend',
            field=models.ForeignKey(to='emails.EmailBackend', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='email',
            name='type',
            field=models.ForeignKey(default=None, blank=True, to='contenttypes.ContentType', null=True),
            preserve_default=True,
        ),
    ]



from __future__ import unicode_literals
import datetime

# Django
from django.contrib import admin
from django.utils import timezone


class OnSpecificDateListFilter(admin.FieldListFilter):

    format = 'YYYY-MM-DD'
    endswith = '__date'

    def __init__(self, field, request, params, model, model_admin, field_path):
        super(OnSpecificDateListFilter, self).__init__(field, request, params, model, model_admin, field_path)

    def expected_parameters(self):
        return [self.field_path + self.endswith]

    def lookups(self, request, model_admin):
        return []

    @property
    def qs_value(self):
        return self.field_path + self.endswith

    def choices(self, cl):
        for x in xrange(31):
            date = timezone.now()
            # By default, offer 15 days ago to 15 days from now
            date = date + datetime.timedelta(days=(x - 15))
            values = {
                self.qs_value: '%s-%s-%s' % (date.year, date.month, date.day,)
            }
            yield {
                'selected': self.used_parameters.get(self.qs_value, None) == values[self.qs_value],
                'query_string': cl.get_query_string(values),
                'display': date.strftime('%a %b %d, %Y')
            }

    def queryset(self, request, queryset):
        processed_params = self.process_used_parameters()
        return queryset.filter(**processed_params)

    def process_used_parameters(self):
        """
        Because we accept a querystring value like "my_field__date=2014-01-01",
        which doesn't make sense by itself to the ORM, we have to do some parsing.
        """
        processed_params = {}
        for key, value in self.used_parameters.items():
            if key.endswith(self.endswith):
                # Split into ``field_name`` and "date" on the "__"
                key_field, key_suffix = key.split('__')

                # Split the date, too
                year, month, day = value.split('-')

                processed_params[key_field + '__year'] = int(year)
                processed_params[key_field + '__month'] = int(month)
                processed_params[key_field + '__day'] = int(day)
            else:
                processed_params[key] = value
        return processed_params

# register the filter
admin.filters.FieldListFilter.register(
    lambda f: isinstance(f, (models.DateField, models.DateTimeField,)), OnSpecificDateListFilter)



from __future__ import unicode_literals

# Django
from django.core.management.base import BaseCommand, CommandError

# Local Apps
from grapevine.sender import ScheduledSendableSender


class Command(BaseCommand):
    help = "Sends or queues all eligible Messages"

    def handle(self, *args, **options):
        sender = ScheduledSendableSender()
        num_sent, num_sendables = sender.deliver_messages()
        self.stdout.write("Sent %s messages from %s total Sendable models." % (num_sent, num_sendables,))



from __future__ import unicode_literals
import re

# Django
from django.shortcuts import render_to_response
from django.template import Template, Context, RequestContext


def render_view(request, template_name, context={}):
    return render_to_response(template_name, context, context_instance=RequestContext(request))


def simple_render(string, context, should_raise=True):
    """
    Renders a stringing against a context. For example, with the
    context: {"key", "value"}, the string "Hello, {{key}}!" would
    be converted to "Hello, value!"
    """
    return Template(string).render(Context(context))



from __future__ import unicode_literals

# Django
from django.contrib import admin

# Local Apps
from .admin_base import BaseModelAdmin
from emails.admin import *
from grapevine.models import QueuedMessage


class QueuedMessageAdmin(BaseModelAdmin):
    list_display = ['id', 'message_type', 'message_id']


admin.site.register(QueuedMessage, QueuedMessageAdmin)



import os
import re
import sys
from setuptools import setup


# Borrowed from the infamous
# https://github.com/tomchristie/django-rest-framework/blob/master/setup.py
def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


# Borrowed from the infamous
# https://github.com/tomchristie/django-rest-framework/blob/master/setup.py
def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


# Borrowed from the infamous
# https://github.com/tomchristie/django-rest-framework/blob/master/setup.py
def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}


version = get_version('grapevine')


if sys.argv[-1] == 'publish-test':
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push origin master")
    os.system("git push --tags")
    os.system("python setup.py sdist upload -r pypitest")
    sys.exit()


CLASSIFIERS = [
    "Framework :: Django",
    "Intended Audience :: Developers",
    "Programming Language :: Python",
]

INSTALL_REQUIRES = [
    "tablets",
    "html2text",
    "requests",
]

setup(
    name="django-grapevine",
    packages=get_packages("grapevine"),
    package_data=get_package_data("grapevine"),
    version=version,
    description="A comprehensive email tool for Django",
    url="https://github.com/craiglabenz/django-grapevine",
    download_url="https://github.com/craiglabenz/django-grapevine/tarball/{0}".format(version),
    keywords=["django", "email", "sending", "tracking"],
    classifiers=CLASSIFIERS,
    install_requires=INSTALL_REQUIRES,
    author="Craig Labenz",
    author_email="craig.labenz@gmail.com",
    license="MIT"
)



# Sys
import json
import requests

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# Django
from django.conf import settings
from django.core.mail.message import sanitize_address

# Local
from grapevine.emails.backends.base import GrapevineEmailBackend


class MailgunAPIError(Exception):
    pass


class EmailBackend(GrapevineEmailBackend):
    """A Django Email backend that uses mailgun. Inspiration: https://github.com/bradwhittington/django-mailgun
    """
    DISPLAY_NAME = "mailgun"
    IMPORT_PATH = "grapevine.emails.backends.MailGunEmailBackend"

    def __init__(self, fail_silently=False, *args, **kwargs):
        access_key, server_name = (kwargs.pop('access_key', None),
                                   kwargs.pop('server_name', None))

        super(EmailBackend, self).__init__(
                        fail_silently=fail_silently,
                        *args, **kwargs)

        try:
            self._access_key = access_key or getattr(settings, 'MAILGUN_ACCESS_KEY', '')
            self._server_name = server_name or getattr(settings, 'MAILGUN_SERVER_NAME', '')
        except AttributeError:
            if fail_silently:
                self._access_key, self._server_name = None
            else:
                raise

        self._api_url = "https://api.mailgun.net/v2/%s/" % self._server_name

    def open(self):
        """Stub for open connection, all sends are done over HTTP POSTs
        """
        pass

    def close(self):
        """Close any open HTTP connections to the API server.
        """
        pass

    def _send(self, email_message):
        """A helper method that does the actual sending."""
        if not email_message.recipients():
            return False

        data = self.prepare_data(email_message)

        try:
            self.r = self.post(email_message, data)
            print self.r.status_code, self.r.content
        except Exception as e:
            if not self.fail_silently:
                raise
            return False

        if self.r.status_code != 200:
            failure_dict = json.loads(self.r.content)
            failure_dict['status'] = self.r.status_code
            self.failure_reason = json.dumps(failure_dict)
            if not self.fail_silently:
                raise MailgunAPIError(self.r)
            return False

        return True

    def prepare_data(self, email_message):
        from_email = sanitize_address(email_message.from_email, email_message.encoding)
        to_recipients = [sanitize_address(addr, email_message.encoding) for addr in email_message.to]
        cc_recipients = [sanitize_address(addr, email_message.encoding) for addr in email_message.cc]
        bcc_recipients = [sanitize_address(addr, email_message.encoding) for addr in email_message.bcc]

        data = {
            "to": ", ".join(to_recipients),
            "from": from_email,
            "subject": email_message.subject,
            "text": email_message.body,
        }

        # # Attach an HTML body if one was set
        # alternatives = getattr(email_message, "alternatives", [])
        # if alternatives:
        #     for content, mimetype in alternatives:
        #         if mimetype == "text/html":
        #             data["html"] = content
        #             break

        if cc_recipients:
            data["cc"] = ", ".join(cc_recipients)

        if bcc_recipients:
            data["bcc"] = ", ".join(bcc_recipients)

        data["v:grapevine-guid"] = email_message._email.guid
        return data

    def post(self, email_message, data):
        return requests.post(self._api_url + "messages.mime",
            auth=("api", self._access_key),
            data=data,
            files={
                "message": StringIO(email_message.message().as_string()),
            }
        )

    def send_messages(self, email_messages):
        """Sends one or more EmailMessage objects and returns the number of
        email messages sent.
        """
        if not email_messages:
            return

        num_sent = 0
        for message in email_messages:
            if self._send(message):
                num_sent += 1

        return num_sent



from __future__ import unicode_literals

# Django
from django.db import models
from django.conf import settings
from django.utils import timezone, six, module_loading
from django.core.mail import get_connection, EmailMultiAlternatives

# 3rd Party
import html2text
# from celery import shared_task

# Local Apps
from utils import parse_email, EventRepo
from managers import EmailManager, EmailRecipientManager
from grapevine.decorators import memoize
from grapevine.models_base import GrapevineModel
from grapevine.models import Transport
from grapevine.settings import grapevine_settings


class Email(Transport):
    """
    The grand central email table.

    Direct Usage:
            # FIRST
        email = Email.objects.create(
                    html_body="<p>Hello, world.</p>",
                    from_email="Support Team <support@acmetech.com>",  * Optional, defaults to settings
                    reply_to="specific-reply-address@acmetech.com",
                    subject="Welcome to Acme Tech!",
                    backend="path.to.backend.EmailBackend", # OPTIONAL.
                    to=[
                        'tos': ["some@email.com", "Frank the Tank <frank@thetank.com", ...],
                        'ccs': ["account-admin1@domain.com", "account-admin2@domain.com", ...]
                        'bccs': "alerts@acmetech.com"  * Could also be a list like above
                    ]
                )

            # THEN....
        email.send()

    """
    backend = models.ForeignKey('EmailBackend', blank=True)
    from_email = models.CharField(max_length=255, db_index=True, default=settings.DEFAULT_FROM_EMAIL)
    reply_to = models.EmailField(max_length=255, blank=True, default=settings.DEFAULT_REPLY_TO)
    subject = models.CharField(max_length=255, default=settings.DEFAULT_SUBJECT)

    objects = EmailManager()

    class Meta:
        app_label = "emails"
        verbose_name = 'Email'
        verbose_name_plural = 'Emails'

    def __unicode__(self):
        recipient = ''

        main_recipient = self.get_main_recipient()
        if main_recipient:
            recipient = ' %s: ' % (main_recipient.__unicode__(),)
        if self.pk:
            return '%s (%s)' % (recipient, self.subject[:50],)
        else:
            return 'Unsaved Email:%s (%s)' % (recipient, self.subject[:50],)

    def get_main_recipient(self):
        if hasattr(self, "main_recipient"):
            return self.main_recipient

        try:
            self.main_recipient = self.recipients.all()[0]
        except IndexError:
            self.main_recipient = None

        return self.main_recipient

    def save(self, *args, **kwargs):
        if not self.backend_id:
            self.backend, created = EmailBackend.objects.get_or_create(path=grapevine_settings.EMAIL_BACKEND)

        super(Email, self).save(*args, **kwargs)

    @staticmethod
    def determine_backend(**kwargs):
        if 'backend_id' in kwargs.keys():
            key = 'backend_id'
            backend = kwargs['backend_id']
        elif 'backend' in kwargs.keys():
            key = 'backend'
            if isinstance(kwargs['backend'], six.string_types):
                backend, created = EmailBackend.objects.get_or_create(path=kwargs['backend'])
            else:
                backend = kwargs['backend']
        else:
            key = backend = None
        return key, backend

    @staticmethod
    def finish_initial_data(sendable, initial_data, **kwargs):
        initial_data.update({
            'from_email': sendable.get_from_email(),
            'reply_to': sendable.get_reply_to(),
            'subject': sendable.get_subject(),
        })

        # Honor a custom EmailBackend request
        key, backend = Email.determine_backend(**kwargs)
        if backend is not None:
            initial_data[key] = backend

        return initial_data

    def add_tos(self, recipients):
        """
        Args:
        @recipients  {list}   A list of email addresses (strings) that should receive ziss email
        """
        # Cast to a list
        if not isinstance(recipients, list):
            recipients = [recipients]

        for recipient in recipients:
            self.add_recipient(recipient, EmailRecipient.TO)

    def add_ccs(self, recipients):
        """
        Args:
        @recipients  {list}   A list of email addresses (strings) that should receive ziss email
        """
        # Cast to a list
        if not isinstance(recipients, list):
            recipients = [recipients]

        for recipient in recipients:
            self.add_recipient(recipient, EmailRecipient.CC)

    def add_bccs(self, recipients):
        """
        Args:
        @recipients  {list}   A list of email addresses (strings) that should receive ziss email
        """
        # Cast to a list
        if not isinstance(recipients, list):
            recipients = [recipients]

        for recipient in recipients:
            self.add_recipient(recipient, EmailRecipient.BCC)

    def add_recipient(self, recipient, recipient_type):
        """
        Args:
        @recipient      {str}   The email to receive this msg
        @recipient_type {int}   Maps to EmailRecipient.TYPES
        """
        return EmailRecipient.objects.create(email=self, address=recipient, type=recipient_type)

    def add_recipients(self, recipients):
        """
        Accepts and saves recipients in a smorgasboard of formats
        """
        if isinstance(recipients, six.string_types):
            return self.add_tos([recipients])
        elif isinstance(recipients, list):
            return self.add_tos(recipients)
        elif isinstance(recipients, dict):
            # Dicts should have keys "to", "cc", and "bcc"
            self.add_tos(recipients.get('to', []))
            self.add_ccs(recipients.get('cc', []))
            self.add_bccs(recipients.get('bcc', []))
        else:
            raise ValueError("`recipients` must be a str, list, or dict")

    def add_variable(self, key, value):
        return EmailVariable.objects.create(email=self, key=key, value=value)

    @property
    def to(self):
        return [er.prepare_for_email() for er in EmailRecipient.objects.filter(email=self, type=EmailRecipient.TO)]

    @property
    def cc(self):
        return [er.prepare_for_email() for er in EmailRecipient.objects.filter(email=self, type=EmailRecipient.CC)]

    @property
    def bcc(self):
        return [er.prepare_for_email() for er in EmailRecipient.objects.filter(email=self, type=EmailRecipient.BCC)]

    def _send(self, backend=None, fail_silently=False, **kwargs):
        """
        Loads up the EmailBackend class to finally talk to an EMAIL_BACKEND
        """
        return self.backend.send(self, fail_silently, **kwargs)

    @property
    @memoize
    def is_spam(self):
        return self.find_event_by_name('Spam Report')

    @property
    @memoize
    def is_opened(self):
        return self.find_event_by_name('Open')

    @property
    @memoize
    def is_clicked(self):
        return self.find_event_by_name('Click')

    @property
    @memoize
    def cached_events(self):
        return [event for event in self.events.all()]

    def find_event_by_name(self, name):
        event_type = EventRepo()[name]
        for email_event in self.cached_events:
            if email_event.event_id == event_type.pk:
                return True
        return False


class EmailRecipient(GrapevineModel):

    TO = 1
    CC = 2
    BCC = 3
    TYPES = (
        (TO, 'To',),
        (CC, 'CC',),
        (BCC, 'BCC',),
    )

    email = models.ForeignKey(Email, db_index=True, related_name='recipients')
    # https://docs.djangoproject.com/en/1.6/ref/models/fields/#emailfield
    address = models.EmailField(max_length=254, db_index=True, verbose_name='Recipient Address')
    domain = models.CharField(max_length=100, db_index=True)
    name = models.CharField(max_length=150, blank=True)
    type = models.IntegerField(choices=TYPES, default=TO)

    objects = EmailRecipientManager()

    class Meta:
        app_label = "emails"
        verbose_name = 'Email Recipient'
        verbose_name_plural = 'Email Recipients'

    def __unicode__(self):
        return '%s: %s' % (self.email.pk, self.prepare_for_email(),)

    def prepare_for_email(self):
        if self.name:
            return '%s <%s>' % (self.name, self.address,)
        else:
            return self.address

    def determine_domain(self):
        if not bool(self.domain):
            self.domain = self.address.split('@')[1].lower()

    def save(self, *args, **kwargs):
        self.determine_domain()
        super(EmailRecipient, self).save(*args, **kwargs)


class EmailBackend(GrapevineModel):
    name = models.CharField(max_length=255)
    path = models.CharField(max_length=255, help_text="The dotted import path to this backend, in "
                               "a structure that would satisfy settings.EMAIL_BACKEND.")
    username = models.CharField(max_length=255, blank=True, help_text="Depending on the provider, this "
                               "may actually be an API key.")
    password = models.CharField(max_length=255, blank=True, help_text="Depending on the provider, this "
                               "may actually be an API secret.")

    class Meta:
        app_label = "emails"
        verbose_name = "Email Backend"
        verbose_name_plural = "Email Backends"

    @property
    def kls(self):
        return module_loading.import_by_path(dotted_path=self.path)

    def __unicode__(self):
        if self.name:
            return self.name
        else:
            return self.path

    def get_connection(self, fail_silently=False, **kwargs):
        """
        Keyword Arguments:
        @backend        {str}   The dotted path to a fully functional Django emailer backend.
                                Defaults to ``settings.EMAIL_BACKEND`` if not passed.
        @fail_silently  {bool}  Whether or not to raise problems.
        """
        if self.username:
            kwargs.setdefault('username', self.username)
        if self.password:
            kwargs.setdefault('password', self.password)

        return get_connection(backend=self.path, fail_silently=fail_silently, **kwargs)

    def get_headers(self, email):
        return {
            'reply_to': email.reply_to,
        }

    def as_message(self, email):
        """
        Django email code relies on a combination of the EMAIL_BACKEND
        class itself and an instance of a message-style class that the
        backend uses to build the actual payload.

        This function is an opportunity for EMAIL_BACKENDs to hook in
        and create a Message object of its own designs.
        """
        import backends
        is_grapevine_backend = issubclass(self.kls, backends.base.GrapevineEmailBackend)
        implements_as_message = hasattr(self.kls, 'as_message')

        if is_grapevine_backend and implements_as_message:
            # Provider-specific EMAIL_BACKENDS will
            # likely implement this method and thus create their
            # own "message" obj
            msg = self.kls.as_message(email)
        else:
            # OTS Django or other SMTP-based EMAIL_BACKENDS will use
            # this standard functionality
            msg = EmailMultiAlternatives(
                subject=email.subject,
                body=email.text_body,
                from_email=email.from_email,
                to=email.to,
                cc=email.cc,
                bcc=email.bcc,
                headers=self.get_headers(email),
            )
            if email.html_body:
                msg.attach_alternative(email.html_body, "text/html")

        msg._email = email
        return msg

    def finalize_message(self, message):
        # Honor unsubscribes and DEV settings
        if hasattr(self.kls, 'finalize_message'):
            return self.kls.finalize_message(message)
        else:
            # The base implementation of this function works for Django's
            # ``EmailMessage`` class
            import backends
            return backends.base.GrapevineEmailBackend.finalize_message(message)

    def send(self, email, fail_silently=False, **kwargs):
        """
        Does a handful of things seen in Django's own wrapper around
        EMAIL_BACKENDS, found at ``django.core.mail.send_mail``.
        """
        kwargs.setdefault('fail_silently', fail_silently)

        # Convert our Grapevine ``Email`` obj into something this
        # specific backend can use
        msg = self.as_message(email)

        # Pass old school EmailMessage instances to
        # the Base finalizer to honor Unsubscribes and such
        msg = self.finalize_message(msg)

        if len(msg.to) == 0:
            self.get_logger().info("Not sending Email Id:%s due to all `TO` recipients having unsubscribed.", email.pk)
            email.status = email.UNSUBSCRIBED
            email.save()
            return False

        # Initialize an instance of the Backend class
        msg.connection = self.get_connection(**kwargs)

        # Fly birdies!
        # http://img3.wikia.nocookie.net/__cb20110615180140/gameofthrones/images/1/18/Bronn_defeats_Vardis.jpg
        is_sent = self._send(msg, **kwargs)

        msg.connection.close()

        if not is_sent:
            if getattr(msg, "failure_reason", False):
                print msg.failure_reason
                email.append_to_log(msg.failure_reason)
            else:
                print "no failure_reason"

        return is_sent

    def _send(self, msg, **kwargs):
        """
        Like ``as_message``, this is a hook for a specific driver to
        optionally take control of a piece of the fun.

        Finally gets bits onto the wire.
        """
        if hasattr(msg.connection, 'send'):
            return msg.connection.send(msg)
        else:
            return msg.send(**kwargs)


class EmailVariable(GrapevineModel):

    email = models.ForeignKey(Email, related_name="variables")
    key = models.CharField(max_length=100)
    value = models.CharField(max_length=5000, blank=True, help_text="1 represents True, and 0 represents False.")

    class Meta:
        app_label = "emails"
        verbose_name = "Email Variable"
        verbose_name_plural = "Email Variables"

        index_together = (
            ('key', 'value',),
        )

    def __unicode__(self):
        if self.value:
            return "%s: %s %s" % (self.email.__unicode__(), self.key, self.value,)
        else:
            return "%s: %s" % (self.email.__unicode__(), self.key,)


class RawEvent(GrapevineModel):

    backend = models.ForeignKey(EmailBackend)
    payload = models.TextField()
    processed_on = models.DateTimeField(null=True, default=None, blank=True,
        verbose_name="Processed On")
    processed_in = models.DecimalField(max_digits=5, decimal_places=4, blank=True,
        null=True, default=None, verbose_name="Processed In")
    is_queued = models.BooleanField(default=False, db_index=True, verbose_name="Is Queued")

    # Email providers do not often send payloads in formats that match
    # their documentation
    is_broken = models.NullBooleanField(default=None, db_index=True, verbose_name="Is Broken")

    # Bookkeeping
    remote_ip = models.GenericIPAddressField(db_index=True, verbose_name="Remote IP")

    class Meta:
        app_label = "emails"
        verbose_name = "Raw Event"
        verbose_name_plural = "Raw Events"

    @staticmethod
    # @shared_task
    def async_process(raw_event_id):
        raw_event = RawEvent.objects.get(pk=raw_event_id)
        raw_event.process()

    def process(self):
        is_processed, time_taken = self.backend.kls().process_event(self)
        # Not queued anymore, obviously
        self.is_queued = False

        if is_processed:
            self.processed_on = timezone.now()
            self.processed_in = time_taken
        else:
            self.is_broken = True
        self.save()


class Event(GrapevineModel):
    name = models.CharField(max_length=100)
    should_stop_sending = models.BooleanField(default=False,
        verbose_name="Should Stop Sending")

    class Meta:
        app_label = "emails"
        verbose_name = "Event"
        verbose_name_plural = "Events"

    def __unicode__(self):
        return self.name


class EmailEvent(GrapevineModel):

    email = models.ForeignKey(Email, related_name="events")
    event = models.ForeignKey(Event, related_name="email_events")
    raw_event = models.ForeignKey(RawEvent, verbose_name="Raw Event", related_name="email_events")
    happened_at = models.DateTimeField(db_index=True, verbose_name="Happened At")

    class Meta:
        app_label = "emails"
        verbose_name = "Email Event"
        verbose_name_plural = "Email Events"

    def __unicode__(self):
        return '%s Email Id: %s' % (self.event.__unicode__(), self.email_id,)


class UnsubscribedAddress(GrapevineModel):
    address = models.CharField(max_length=255, db_index=True)
    email = models.ForeignKey(Email, null=True, blank=True, default=None,
        help_text="Optional link back to the Email in which this recipient address \
        clicked Unsubscribe.", related_name="unsubscribed_addresses")

    class Meta:
        app_label = "emails"
        verbose_name = "Unsubscribed Address"
        verbose_name_plural = "Unsubscribed Addresses"

    def __unicode__(self):
        return self.address

    def save(self, *args, **kwargs):
        """
        Parses out formatted email addresses into raw thangs
        """
        name, self.address = parse_email(self.address)
        return super(UnsubscribedAddress, self).save(*args, **kwargs)



from __future__ import unicode_literals
import datetime

# Django
from django.conf.urls import url
from django.core.mail.backends.base import BaseEmailBackend
from django.http import HttpResponse
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt

# Local Apps
from grapevine.settings import grapevine_settings
from grapevine.emails import models as email_models
from grapevine.emails.utils import parse_email


class GrapevineEmailBackend(BaseEmailBackend):
    # Used to register a callback url for the 3rd party
    DISPLAY_NAME = None

    # Used to process catalog events
    IMPORT_PATH = None

    LISTENS_FOR_EVENTS = True

    def get_urls(self):
        urls = []
        if self.LISTENS_FOR_EVENTS:
            urls.append(
                url(r'^backends/{0}/events/$'.format(self.DISPLAY_NAME), self.events_webhook, name="{0}-events-webhook".format(self.DISPLAY_NAME)),
            )
        return urls

    def process_event(self, event):
        """
        This method must only be implemented if this backend
        communicates with a specific email sending provider that
        offers webhooks for email events.
        """
        raise NotImplementedError()

    def datetime_from_seconds(self, timestamp):
        """
        Takes a timestamp (like 1337966815) and turns it into
        something like this: datetime.datetime(2012, 5, 16, 15, 46, 40, tzinfo=<UTC>)
        """
        datetime_obj = datetime.datetime.utcfromtimestamp(timestamp)
        return timezone.make_aware(datetime_obj, timezone.get_default_timezone())

    @classmethod
    def debugify_message(cls, message):
        message.to = [grapevine_settings.DEBUG_EMAIL_ADDRESS]
        message.cc = []
        message.bcc = []

        return message

    @classmethod
    def finalize_message(cls, message):
        """
        Where the ``DEBUG`` setting and unsubscribe list are honored.

        Note that this works for the Django ``EmailMessage`` class as well
        as the SendGrid ``mail.Mail()`` class, as the two store recipients
        internally the same way.
        """
        # Strip out all recipients if this is a (localhost-style)
        # TEST_EMAIL.
        if grapevine_settings.DEBUG:
            message = cls.debugify_message(message)

        # Compile a grand list of all recipient lists
        recipient_lists = [getattr(message, 'to', []), getattr(message, 'cc', []), getattr(message, 'bcc', [])]
        for recipient_list in recipient_lists:
            indicies_to_pop = []
            for index, recipient in enumerate(recipient_list):
                try:
                    # Remove any remaining unsubscribed emails
                    name, address = parse_email(recipient)
                    unsubscribed = email_models.UnsubscribedAddress.objects.get(
                        address=address
                    )
                    indicies_to_pop.append(index)
                except email_models.UnsubscribedAddress.DoesNotExist:
                    pass

            # Must loop over this list in reverse to not change
            # the indicies of elements after the one's we remove
            indicies_to_pop.reverse()
            for index in indicies_to_pop:
                recipient_list.pop(index)

        return message

    @csrf_exempt
    def events_webhook(self, request):
        """
        Responds to {POST /grapevine/backends/:DISPLAY_NAME/events/}
        """
        # Using the ``require_POST`` decorator causes problems
        if request.method != 'POST':
            return HttpResponse(status=405)

        try:
            backend = email_models.EmailBackend.objects.filter(path=self.IMPORT_PATH)[0]
        except IndexError:
            backend = email_models.EmailBackend.objects.create(path=self.IMPORT_PATH)

        # Pull out the payload from request.POST as per
        # https://docs.djangoproject.com/en/1.6/ref/request-response/#django.http.HttpRequest.POST
        email_models.RawEvent.objects.create(backend=backend, payload=request.body,
                                             remote_ip=request.META['REMOTE_ADDR'])

        # A 200 tells SendGrid we successfully accepted this event payload
        return HttpResponse(status=200)



from __future__ import unicode_literals

# Django
from django.utils import timezone
from django.db.models.query import QuerySet

# Local Apps
from . import models


class SendableQuerySet(QuerySet):

    def sent(self):
        return self.exclude(message_id=None)

    def unsent(self):
        return self.filter(message_id=None, cancelled_at_send_time=False)

    def ready_to_send(self):
        return self.filter(scheduled_send_time__lte=timezone.now())

    def not_queued(self):
        message_ids = models.QueuedMessage.objects.\
            filter(message_type=self.model.content_type).\
            values_list('message_id', flat=True)

        return self.exclude(pk__in=message_ids)

    def is_sendable(self):
        """
        Override hook for child querysets to amend logic to remove
        scheduled emails that have since become obselete.
        """
        return self

    def is_eligible(self):
        return self.unsent().ready_to_send().not_queued().is_sendable()

    def with_messages(self, should_preload_events=True):
        """
        Alert: Evaluates the queryset and attaches the ``_message`` attr!
        **DOES NOT SUPPORT CHAINING**
        """
        objs = []

        # Email, HipChat, whatever, just load the relevant class
        transport_class = self.model.get_transport_class()

        # Distill the message_ids as a map back to their objs
        message_id_map = {}
        for obj in self:
            message_id_map[obj.message_id] = obj
            objs.append(obj)

        # Load all relevant messages
        messages = transport_class._default_manager.filter(pk__in=message_id_map.keys())

        if should_preload_events and hasattr(transport_class, "events"):
            messages = messages.prefetch_related("events")

        for message in messages:
            message_id_map[message.pk]._message = message

        return objs



from __future__ import unicode_literals

# Django
from django import forms
from django.contrib.admin.widgets import AdminSplitDateTime

# Local Apps
from tablets.models import Template


class AdminBulkEditForm(forms.Form):

    template = forms.ModelChoiceField(queryset=Template.objects.all(), help_text='Supply a value for this OR the Frozen template.')
    customized_from_email = forms.CharField(label='Frozen From Email')
    customized_reply_to = forms.CharField(label='Frozen Reply To')
    customized_subject = forms.CharField(label='Frozen Subject')
    customized_template = forms.CharField(label='Frozen Template', widget=forms.Textarea, help_text='A value here will trump anything supplied for "Template".')
    scheduled_send_time = forms.DateTimeField(label='Send Time', widget=AdminSplitDateTime)
